@@ -52,7 +52,7 @@ static IDirectFBDisplayLayer *layer       = NULL;
 static IDirectFBSurface      *arrow       = NULL;
 static IDirectFBEventBuffer  *events      = NULL;
 static DFBSurfacePixelFormat  pixelformat = DSPF_UNKNOWN;
-static int                    event_timer = -1;
+static struct timer *event_timer = NULL;
 
 
 static inline void directfb_set_color  (IDirectFBSurface *surface, long color);
@@ -68,7 +68,7 @@ static struct graphics_device * directfb_lookup_in_table (DFBWindowID  id);
 
 
 static unsigned char *
-directfb_fb_init_driver (unsigned char *param, unsigned char *display)
+directfb_init_driver (unsigned char *param, unsigned char *display)
 {
   DFBDisplayLayerConfig  config;
   DFBResult              ret;
@@ -439,12 +439,16 @@ directfb_vscroll (struct graphics_device *gd, struct rect_set **set, int sc)
 struct graphics_driver directfb_driver =
 {
   cast_uchar "directfb",
-  directfb_fb_init_driver,
+  directfb_init_driver,
   directfb_init_device,
   directfb_shutdown_device,
   directfb_shutdown_driver,
-  dummy_emergency_shutdown,
+  NULL,
+  NULL,
   directfb_get_driver_param,
+  NULL,
+  NULL,
+  NULL,
   directfb_get_empty_bitmap,
   directfb_register_bitmap,
   directfb_prepare_strip,
@@ -748,7 +752,7 @@ directfb_add_to_table (struct graphics_device *gd)
   int i;
 
   i = data->id % DIRECTFB_HASH_TABLE_SIZE;
- 
+
   devices = directfb_hash_table[i];
 
   if (devices)
