@@ -90,13 +90,9 @@ int dummy_unblock(struct graphics_device *dev)
 	return 0;
 }
 
-void dummy_emergency_shutdown(void)
-{
-}
-
 #if 0
 static unsigned char *list_graphics_drivers(void)
-{ 
+{
 	unsigned char *d = init_str();
 	int l = 0;
 	struct graphics_driver **gd;
@@ -191,6 +187,7 @@ void shutdown_graphics(void)
 		if (drv->shell) mem_free(drv->shell);
 		drv->shutdown_driver();
 		drv = NULL;
+		F = 0;
 	}
 }
 
@@ -223,12 +220,11 @@ void generic_set_clip_area(struct graphics_device *dev, struct rect *r)
 
 #if defined(GRDRV_SVGALIB) || defined(GRDRV_FB) || defined(GRDRV_GRX)
 
-static struct graphics_driver *virtual_device_driver;
 struct graphics_device **virtual_devices;
 int n_virtual_devices = 0;
 struct graphics_device *current_virtual_device;
 
-static int virtual_device_timer;
+static struct timer *virtual_device_timer;
 
 int init_virtual_devices(struct graphics_driver *drv, int n)
 {
@@ -239,8 +235,7 @@ int init_virtual_devices(struct graphics_driver *drv, int n)
 	if ((unsigned)n > MAXINT / sizeof(struct graphics_device *)) overalloc();
 	virtual_devices = mem_calloc(n * sizeof(struct graphics_device *));
 	n_virtual_devices = n;
-	virtual_device_driver = drv;
-	virtual_device_timer = -1;
+	virtual_device_timer = NULL;
 	current_virtual_device = NULL;
 	return 0;
 }
@@ -251,10 +246,10 @@ struct graphics_device *init_virtual_device(void)
 	for (i = 0; i < n_virtual_devices; i++) if (!virtual_devices[i]) {
 		struct graphics_device *dev;
 		dev = mem_calloc(sizeof(struct graphics_device));
-		dev->size.x2 = virtual_device_driver->x;
-		dev->size.y2 = virtual_device_driver->y;
+		dev->size.x2 = drv->x;
+		dev->size.y2 = drv->y;
 		current_virtual_device = virtual_devices[i] = dev;
-		virtual_device_driver->set_clip_area(dev, &dev->size);
+		drv->set_clip_area(dev, &dev->size);
 		return dev;
 	}
 	return NULL;
@@ -262,7 +257,7 @@ struct graphics_device *init_virtual_device(void)
 
 static void virtual_device_timer_fn(void *p)
 {
-	virtual_device_timer = -1;
+	virtual_device_timer = NULL;
 	if (current_virtual_device && current_virtual_device->redraw_handler) {
 		drv->set_clip_area(current_virtual_device, &current_virtual_device->size);
 		current_virtual_device->redraw_handler(current_virtual_device, &current_virtual_device->size);
@@ -285,7 +280,7 @@ void switch_virtual_device(int i)
 	if (i < 0 || i >= n_virtual_devices || !virtual_devices[i]) return;
 	current_virtual_device = virtual_devices[i];
 	ok_switch:
-	if (virtual_device_timer == -1)
+	if (virtual_device_timer == NULL)
 		virtual_device_timer = install_timer(0, virtual_device_timer_fn, NULL);
 }
 
@@ -311,6 +306,21 @@ void shutdown_virtual_device(struct graphics_device *dev)
 	/*internal("shutdown_virtual_device: device not initialized");*/
 }
 
+void resize_virtual_devices(int x, int y)
+{
+	int i;
+	drv->x = x;
+	drv->y = y;
+	for (i = 0; i < n_virtual_devices; i++) {
+		struct graphics_device *dev = virtual_devices[i];
+		if (dev) {
+			dev->size.x2 = x;
+			dev->size.y2 = y;
+			dev->resize_handler(dev);
+		}
+	}
+}
+
 void shutdown_virtual_devices(void)
 {
 	int i;
@@ -321,7 +331,7 @@ void shutdown_virtual_devices(void)
 	for (i = 0; i < n_virtual_devices; i++) if (virtual_devices[i]) internal("shutdown_virtual_devices: virtual device %d is still active", i);
 	mem_free(virtual_devices);
 	n_virtual_devices = 0;
-	if (virtual_device_timer != -1) kill_timer(virtual_device_timer), virtual_device_timer = -1;
+	if (virtual_device_timer != NULL) kill_timer(virtual_device_timer), virtual_device_timer = NULL;
 }
 
 #endif
