@@ -144,7 +144,7 @@ static int check_http_server_bugs(unsigned char *url, struct http_connection_inf
 		mem_free(server);
 		return bugs & ~BL_NO_RANGE;
 	}
-	return 0;	
+	return 0;
 }
 
 static void http_end_request(struct connection *c, int notrunc, int nokeepalive, int state)
@@ -205,7 +205,7 @@ static void add_url_to_str(unsigned char **str, int *l, unsigned char *url)
 static void http_send_header(struct connection *c)
 {
 	struct http_connection_info *info;
-	int http10 = http_options.http10;
+	int http10 = http_options.http10 && !SCRUB_HEADERS;
 	int proxy;
 	unsigned char *hdr;
 	unsigned char *h, *u;
@@ -220,7 +220,7 @@ static void http_send_header(struct connection *c)
 	}
 
 	proxy = upcase(c->url[0]) == 'P';
-	host = !proxy ? c->url : get_url_data(c->url);
+	host = remove_proxy_prefix(c->url);
 	set_connection_timeout(c);
 	info = mem_calloc(sizeof(struct http_connection_info));
 	c->info = info;
@@ -228,11 +228,7 @@ static void http_send_header(struct connection *c)
 	info->https_forward = !c->ssl && proxy && host && !casecmp(host, cast_uchar "https://", 8);
 	if (c->ssl) proxy = 0;
 #endif
-	info->send_close = info->https_forward || http10 || (post && http_options.bug_post_no_keepalive)
-#ifdef HAVE_SSL
-		|| c->ssl
-#endif
-		;
+	info->send_close = info->https_forward || http10 || (post && http_options.bug_post_no_keepalive);
 	hdr = init_str();
 	if (!host) {
 		http_bad_url:
@@ -354,7 +350,9 @@ static void http_send_header(struct connection *c)
 static void add_user_agent(unsigned char **hdr, int *l)
 {
 	add_to_str(hdr, l, cast_uchar "User-Agent: ");
-	if (!(*http_options.header.fake_useragent)) {
+	if (SCRUB_HEADERS) {
+		add_to_str(hdr, l, cast_uchar "Mozilla/5.0 (Windows NT 6.1; rv:31.0) Gecko/20100101 Firefox/31.0\r\n");
+	} else if (!(*http_options.header.fake_useragent)) {
 		add_to_str(hdr, l, cast_uchar("Links (" VERSION_STRING "; "));
 		add_to_str(hdr, l, system_name);
 		add_to_str(hdr, l, cast_uchar "; ");
@@ -383,6 +381,8 @@ static void add_user_agent(unsigned char **hdr, int *l)
 
 static void add_referer(unsigned char **hdr, int *l, unsigned char *url, unsigned char *prev_url)
 {
+	if (SCRUB_HEADERS)
+		return;
 	switch (http_options.header.referer)
 	{
 		case REFERER_FAKE:
@@ -390,7 +390,7 @@ static void add_referer(unsigned char **hdr, int *l, unsigned char *url, unsigne
 		add_to_str(hdr, l, http_options.header.fake_referer);
 		add_to_str(hdr, l, cast_uchar "\r\n");
 		break;
-		
+
 		case REFERER_SAME_URL:
 		add_to_str(hdr, l, cast_uchar "Referer: ");
 		add_url_to_str(hdr, l, url);
@@ -404,6 +404,12 @@ static void add_referer(unsigned char **hdr, int *l, unsigned char *url, unsigne
 			if ((h = get_host_name(url))) {
 				if ((j = get_host_name(prev_url))) {
 					if (!strcasecmp(cast_const_char h, cast_const_char j)) brk = 0;
+					else if (!strcasecmp(cast_const_char h, "imageproxy.jxs.cz")) {
+						int l = (int)strlen(cast_const_char j);
+						int q = (int)strlen(".blog.cz");
+						if (l > q && !strcasecmp(cast_const_char (j + l - q), ".blog.cz")) brk = 0;
+						else if (!strcasecmp(cast_const_char j, "blog.cz")) brk = 0;
+					}
 					mem_free(j);
 				}
 				mem_free(h);
@@ -433,20 +439,28 @@ static void add_referer(unsigned char **hdr, int *l, unsigned char *url, unsigne
 
 static void add_accept(unsigned char **hdr, int *l)
 {
-	add_to_str(hdr, l, cast_uchar "Accept: */*\r\n");
+	if (SCRUB_HEADERS) {
+		add_to_str(hdr, l, cast_uchar "Accept: text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8\r\n");
+	} else {
+		add_to_str(hdr, l, cast_uchar "Accept: */*\r\n");
+	}
 }
 
 static void add_accept_language(unsigned char **hdr, int *l, struct http_connection_info *info)
 {
 	if (!(info->bl_flags & BL_NO_ACCEPT_LANGUAGE)) {
-		int la;
 		add_to_str(hdr, l, cast_uchar "Accept-Language: ");
-		la = *l;
-		add_to_str(hdr, l, _(TEXT_(T__ACCEPT_LANGUAGE), NULL));
-		add_to_str(hdr, l, cast_uchar ",");
-		if (!strstr(cast_const_char(*hdr + la), "en,") &&
-		    !strstr(cast_const_char(*hdr + la), "en;")) add_to_str(hdr, l, cast_uchar "en;q=0.2,");
-		add_to_str(hdr, l, cast_uchar "*;q=0.1\r\n");
+		if (SCRUB_HEADERS) {
+			add_to_str(hdr, l, cast_uchar "en-us,en;q=0.5\r\n");
+		} else {
+			int la;
+			la = *l;
+			add_to_str(hdr, l, _(TEXT_(T__ACCEPT_LANGUAGE), NULL));
+			add_to_str(hdr, l, cast_uchar ",");
+			if (!strstr(cast_const_char(*hdr + la), "en,") &&
+			    !strstr(cast_const_char(*hdr + la), "en;")) add_to_str(hdr, l, cast_uchar "en;q=0.2,");
+			add_to_str(hdr, l, cast_uchar "*;q=0.1\r\n");
+		}
 	}
 }
 
@@ -480,20 +494,20 @@ static void add_accept_encoding(unsigned char **hdr, int *l, unsigned char *url,
 		add_to_str(hdr, l, cast_uchar "Accept-Encoding: ");
 		l1 = *l;
 #if defined(HAVE_ZLIB)
-		if (*l != l1) add_chr_to_str(hdr, l, ',');
-		add_to_str(hdr, l, cast_uchar "gzip,deflate");
+		if (*l != l1) add_to_str(hdr, l, cast_uchar ", ");
+		add_to_str(hdr, l, cast_uchar "gzip, deflate");
 #endif
 #if defined(HAVE_BZIP2)
-		if (!(info->bl_flags & BL_NO_BZIP2)) {
-			if (*l != l1) add_chr_to_str(hdr, l, ',');
+		if (!SCRUB_HEADERS && !(info->bl_flags & BL_NO_BZIP2)) {
+			if (*l != l1) add_to_str(hdr, l, cast_uchar ", ");
 			add_to_str(hdr, l, cast_uchar "bzip2");
 		}
 #endif
 	/* LZMA on DOS often fails with out of memory, don't announce it */
 #if defined(HAVE_LZMA) && !defined(DOS)
-		if (!(info->bl_flags & BL_NO_BZIP2)) {
-			if (*l != l1) add_chr_to_str(hdr, l, ',');
-			add_to_str(hdr, l, cast_uchar "lzma,lzma2");
+		if (!SCRUB_HEADERS && !(info->bl_flags & BL_NO_BZIP2)) {
+			if (*l != l1) add_to_str(hdr, l, cast_uchar ", ");
+			add_to_str(hdr, l, cast_uchar "lzma, lzma2");
 		}
 #endif
 		if (*l != l1) add_to_str(hdr, l, cast_uchar "\r\n");
@@ -506,6 +520,8 @@ static void add_accept_encoding(unsigned char **hdr, int *l, unsigned char *url,
 static void add_accept_charset(unsigned char **hdr, int *l, struct http_connection_info *info)
 {
 	static unsigned char *accept_charset = NULL;
+	if (SCRUB_HEADERS)
+		return;
 	if (!accept_charset) {
 		int i;
 		unsigned char *cs, *ac;
@@ -774,7 +790,7 @@ static void read_http_data(struct connection *c, struct read_buffer *rb)
 			if (!info->chunk_remaining && rb->len >= 1) {
 				if (rb->data[0] == 10) kill_buffer_data(rb, 1);
 				else {
-					if (rb->data[0] != 13 || (rb->len >= 2 && rb->data[1] != 10)) {
+					if (rb->data[0] != 13 || (rb->len >= 2 && ((unsigned char *)rb->data)[1] != 10)) {
 						setcstate(c, S_HTTP_ERROR);
 						abort_connection(c);
 						return;
@@ -786,7 +802,6 @@ static void read_http_data(struct connection *c, struct read_buffer *rb)
 				goto next_chunk;
 			}
 		}
-				
 	}
 	read_more:
 	read_from_socket(c, c->sock1, rb, read_http_data);
@@ -799,11 +814,11 @@ static int get_header(struct read_buffer *rb)
 	if (rb->len <= 0) return 0;
 	if (rb->data[0] != 'H') return -2;
 	if (rb->len <= 1) return 0;
-	if (rb->data[1] != 'T') return -2;
+	if (((unsigned char *)rb->data)[1] != 'T') return -2;
 	if (rb->len <= 2) return 0;
-	if (rb->data[2] != 'T') return -2;
+	if (((unsigned char *)rb->data)[2] != 'T') return -2;
 	if (rb->len <= 3) return 0;
-	if (rb->data[3] != 'P') return -2;
+	if (((unsigned char *)rb->data)[3] != 'P') return -2;
 	for (i = 0; i < rb->len; i++) {
 		unsigned char a = rb->data[i];
 		if (/*a < ' ' && a != 10 && a != 13*/ !a) return -1;
@@ -830,7 +845,7 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 	struct cache_entry *e;
 	int previous_http_code;
 	struct http_connection_info *info;
-	unsigned char *host = upcase(c->url[0]) != 'P' ? c->url : get_url_data(c->url);
+	unsigned char *host = remove_proxy_prefix(c->url);
 	set_connection_timeout(c);
 	info = c->info;
 	if (rb->close == 2) {
@@ -881,7 +896,7 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 	}
 	ch = head;
 	while ((cookie = parse_http_header(ch, cast_uchar "Set-Cookie", &ch))) {
-		unsigned char *host = upcase(c->url[0]) != 'P' ? c->url : get_url_data(c->url);
+		unsigned char *host = remove_proxy_prefix(c->url);
 		set_cookie(NULL, host, cookie);
 		mem_free(cookie);
 	}
