@@ -27,9 +27,14 @@
 #endif /* #ifndef C_BIG_ENDIAN */
 #endif /* #ifndef C_LITTLE_ENDIAN */
 #endif /* #ifndef REPACK_16 */
-	  
+
 /* Decoder structs */
 
+struct png_decoder{
+	png_structp png_ptr;
+	png_infop info_ptr;
+};
+
 /* Warning for from-web PNG images */
 static void img_my_png_warning(png_structp a, png_const_charp b)
 {
@@ -128,7 +133,7 @@ static void a2char_from_unsigned_short(unsigned char *chr, unsigned short *shrt,
 static void unsigned_short_from_2char(unsigned short *shrt, unsigned char *chr, int len)
 {
 	unsigned short s;
-	
+
 	for (;len;len--,shrt++,chr+=2){
 		s=((*chr)<<8)|chr[1];
 		*shrt=s;
@@ -249,6 +254,12 @@ void png_restart(struct cached_image *cimg, unsigned char *data, int length)
 	if (end_callback_hit) img_end(cimg);
 }
 
+void png_destroy_decoder(struct cached_image *cimg)
+{
+	struct png_decoder *decoder = (struct png_decoder *)cimg->decoder;
+	png_destroy_read_struct(&decoder->png_ptr, &decoder->info_ptr, NULL);
+}
+
 void add_png_version(unsigned char **s, int *l)
 {
 	add_to_str(s, l, cast_uchar "PNG (");
