@@ -73,23 +73,43 @@ unsigned char get_attribute(int fg, int bg)
 
 struct list_head terminals = {&terminals, &terminals};
 
-static void alloc_term_screen(struct terminal *term, int x, int y)
+static void set_margin(struct terminal *term)
+{
+	struct term_spec *ts = term->spec;
+	if (term->real_x <= ts->left_margin + ts->right_margin) {
+		term->left_margin = 0;
+		term->x = term->real_x;
+	} else {
+		term->left_margin = ts->left_margin;
+		term->x = term->real_x - (ts->left_margin + ts->right_margin);
+	}
+	if (term->real_y <= ts->top_margin + ts->bottom_margin) {
+		term->top_margin = 0;
+		term->y = term->real_y;
+	} else {
+		term->top_margin = ts->top_margin;
+		term->y = term->real_y - (ts->top_margin + ts->bottom_margin);
+	}
+}
+
+
+static void alloc_term_screen(struct terminal *term)
 {
 	chr *s, *t;
 	NO_GFX;
-	if (x < 0) x = 1;
-	if (y < 0) y = 1;
-	if (x && (unsigned)x * (unsigned)y / (unsigned)x != (unsigned)y) overalloc();
-	if ((unsigned)x * (unsigned)y > MAXINT / sizeof(*term->screen)) overalloc();
-	s = mem_realloc(term->screen, x * y * sizeof(*term->screen));
-	t = mem_realloc(term->last_screen, x * y * sizeof(*term->screen));
-	memset(t, -1, x * y * sizeof(*term->screen));
-	term->x = x;
-	term->y = y;
+	if (term->x < 0) term->x = 1;
+	if (term->y < 0) term->y = 1;
+	if (term->x && (unsigned)term->x * (unsigned)term->y / (unsigned)term->x != (unsigned)term->y) overalloc();
+	if ((unsigned)term->x * (unsigned)term->y > MAXINT / sizeof(*term->screen)) overalloc();
+	s = mem_realloc(term->screen, term->x * term->y * sizeof(*term->screen));
+	t = mem_realloc(term->last_screen, term->x * term->y * sizeof(*term->screen));
+	memset(t, -1, term->x * term->y * sizeof(*term->screen));
 	term->last_screen = t;
-	memset(s, 0, x * y * sizeof(*term->screen));
+	memset(s, 0, term->x * term->y * sizeof(*term->screen));
 	term->screen = s;
 	term->dirty = 1;
+	term->lcx = -1;
+	term->lcy = -1;
 }
 
 static void clear_terminal(struct terminal *term)
@@ -104,7 +124,7 @@ void redraw_below_window(struct window *win)
 	int tr;
 	struct terminal *term = win->term;
 	struct window *end = win;
-	struct event ev = { EV_REDRAW, 0, 0, 0 };
+	struct links_event ev = { EV_REDRAW, 0, 0, 0 };
 	NO_GFX;
 	ev.x = term->x;
 	ev.y = term->y;
@@ -120,7 +140,7 @@ void redraw_below_window(struct window *win)
 static void redraw_terminal_ev(struct terminal *term, int e)
 {
 	struct window *win;
-	struct event ev = {0, 0, 0, 0};
+	struct links_event ev = {0, 0, 0, 0};
 	NO_GFX;
 	ev.ev = e;
 	ev.x = term->x;
@@ -157,7 +177,8 @@ static void redraw_terminal_cls(struct terminal *term)
 {
 	NO_GFX;
 	erase_screen(term);
-	alloc_term_screen(term, term->x, term->y);
+	set_margin(term);
+	alloc_term_screen(term);
 	redraw_terminal_all(term);
 }
 
@@ -312,7 +333,7 @@ void draw_to_window(struct window *win, void (*fn)(struct terminal *term, void *
 		if (win->prev == end || term->redrawing) return;
 		term->redrawing = 1;
 		{
-			struct event ev = { EV_REDRAW, 0, 0, 0 };
+			struct links_event ev = { EV_REDRAW, 0, 0, 0 };
 			ev.x = term->x;
 			ev.y = term->y;
 			for (win = win->prev; win != end; win = win->prev) win->handler(win, &ev, 0);
@@ -362,7 +383,7 @@ static void redraw_windows(struct terminal *term)
 	return;
 	ok:
 	foreach(win, term->windows) {
-		struct event ev = { EV_REDRAW, 0, 0, 0 };
+		struct links_event ev = { EV_REDRAW, 0, 0, 0 };
 		ev.x = term->x;
 		ev.y = term->y;
 		drv->set_clip_area(term->dev, &win->redr);
@@ -388,9 +409,9 @@ void set_window_pos(struct window *win, int x1, int y1, int x2, int y2)
 
 #endif
 
-void add_window_at_pos(struct terminal *term, void (*handler)(struct window *, struct event *, int), void *data, struct window *at)
+void add_window_at_pos(struct terminal *term, void (*handler)(struct window *, struct links_event *, int), void *data, struct window *at)
 {
-	struct event ev = { EV_INIT, 0, 0, 0 };
+	struct links_event ev = { EV_INIT, 0, 0, 0 };
 	struct window *win;
 	ev.x = term->x;
 	ev.y = term->y;
@@ -403,14 +424,14 @@ void add_window_at_pos(struct terminal *term, void (*handler)(struct window *, s
 	win->handler(win, &ev, 0);
 }
 
-void add_window(struct terminal *term, void (*handler)(struct window *, struct event *, int), void *data)
+void add_window(struct terminal *term, void (*handler)(struct window *, struct links_event *, int), void *data)
 {
 	add_window_at_pos(term, handler, data, (struct window *)(void *)&term->windows);
 }
 
 void delete_window(struct window *win)
 {
-	struct event ev = { EV_ABORT, 0, 0, 0 };
+	struct links_event ev = { EV_ABORT, 0, 0, 0 };
 	win->handler(win, &ev, 1);
 	del_from_list(win);
 	if (win->data) mem_free(win->data);
@@ -425,7 +446,7 @@ void delete_window(struct window *win)
 	mem_free(win);
 }
 
-void delete_window_ev(struct window *win, struct event *ev)
+void delete_window_ev(struct window *win, struct links_event *ev)
 {
 	struct window *w = win->next;
 	if ((void *)w == &win->term->windows) w = NULL;
@@ -462,7 +483,7 @@ struct ewd {
 	int b;
 };
 
-static void empty_window_handler(struct window *win, struct event *ev, int fwd)
+static void empty_window_handler(struct window *win, struct links_event *ev, int fwd)
 {
 	struct window *n;
 	struct ewd *ewd = win->data;
@@ -505,7 +526,7 @@ void free_term_specs(void)
 
 struct list_head term_specs = {&term_specs, &term_specs};
 
-static struct term_spec dumb_term = { NULL, NULL, "", 0, 1, 0, 0, 0, 0, 0 };
+static struct term_spec dumb_term = { NULL, NULL, "", 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
 
 static struct term_spec *get_term_spec(unsigned char *term)
 {
@@ -515,6 +536,12 @@ static struct term_spec *get_term_spec(unsigned char *term)
 	return &dumb_term;
 }
 
+static void sync_term_specs(void)
+{
+	struct terminal *term;
+	foreach(term, terminals) term->spec = get_term_spec(term->term);
+}
+
 struct term_spec *new_term_spec(unsigned char *term)
 {
 	struct term_spec *t;
@@ -528,13 +555,7 @@ struct term_spec *new_term_spec(unsigned char *term)
 	return t;
 }
 
-void sync_term_specs(void)
-{
-	struct terminal *term;
-	foreach(term, terminals) term->spec = get_term_spec(term->term);
-}
-
-struct terminal *init_term(int fdin, int fdout, void (*root_window)(struct window *, struct event *, int))
+struct terminal *init_term(int fdin, int fdout, void (*root_window)(struct window *, struct links_event *, int))
 {
 	static tcount terminal_count = 0;
 	struct terminal *term;
@@ -554,17 +575,17 @@ struct terminal *init_term(int fdin, int fdout, void (*root_window)(struct windo
 	term->spec = &dumb_term;
 	term->input_queue = DUMMY;
 	init_list(term->windows);
+	term->handle_to_close = -1;
 	win = mem_calloc(sizeof(struct window));
 	win->handler = root_window;
 	win->term = term;
 	add_to_list(term->windows, win);
-	/*alloc_term_screen(term, 80, 25);*/
 	add_to_list(terminals, term);
-	set_handlers(fdin, (void (*)(void *))in_term, NULL, (void (*)(void *))destroy_terminal, term);
+	set_handlers(fdin, (void (*)(void *))in_term, NULL, term);
 	return term;
 }
 
-static int process_utf_8(struct terminal *term, struct event *ev)
+static int process_utf_8(struct terminal *term, struct links_event *ev)
 {
 #if defined(G) || defined(ENABLE_UTF8)
 	if (ev->ev == EV_KBD && ((!F && term->spec->charset == utf8_table)
@@ -593,15 +614,14 @@ static int process_utf_8(struct terminal *term, struct event *ev)
 
 #ifdef G
 
-static struct term_spec gfx_term = { NULL, NULL, "", 0, 0, 0, 0, 0, 0, 0 };
+static struct term_spec gfx_term = { NULL, NULL, "", 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
 
-struct terminal *init_gfx_term(void (*root_window)(struct window *, struct event *, int), void *info, int len)
+struct terminal *init_gfx_term(void (*root_window)(struct window *, struct links_event *, int), unsigned char *cwd, void *info, int len)
 {
 	static tcount terminal_count = 0;
 	struct terminal *term;
 	struct graphics_device *dev;
 	struct window *win;
-	unsigned char *cwd;
 	NO_TXT;
 	term = mem_calloc(sizeof(struct terminal));
 	term->count = terminal_count++;
@@ -620,13 +640,11 @@ struct terminal *init_gfx_term(void (*root_window)(struct window *, struct event
 	term->environment = !(drv->flags & GD_ONLY_1_WINDOW) ? ENV_G : 0;
 	if (!strcasecmp(cast_const_char drv->name, "x")) term->environment |= ENV_XWIN;
 	term->spec = &gfx_term;
-	if ((cwd = get_cwd())) {
-		safe_strncpy(term->cwd, cwd, MAX_CWD_LEN);
-		mem_free(cwd);
-	}
+	safe_strncpy(term->cwd, cwd, MAX_CWD_LEN);
 	gfx_term.charset = utf8_table;
 	if (gfx_term.charset == -1) gfx_term.charset = 0;
 	init_list(term->windows);
+	term->handle_to_close = -1;
 	win = mem_calloc(sizeof (struct window));
 	win->handler = root_window;
 	win->term = term;
@@ -640,7 +658,7 @@ struct terminal *init_gfx_term(void (*root_window)(struct window *, struct event
 	dev->mouse_handler = t_mouse;
 	{
 		int *ptr;
-		struct event ev = { EV_INIT, 0, 0, 0 };
+		struct links_event ev = { EV_INIT, 0, 0, 0 };
 		ev.x = dev->size.x2;
 		ev.y = dev->size.y2;
 		if ((unsigned)len > MAXINT - sizeof(int)) overalloc();
@@ -668,7 +686,7 @@ void t_resize(struct graphics_device *dev)
 {
 	struct terminal *term = dev->user_data;
 	struct window *win;
-	struct event ev = { EV_RESIZE, 0, 0, 0 };
+	struct links_event ev = { EV_RESIZE, 0, 0, 0 };
 	term->x = ev.x = dev->size.x2;
 	term->y = ev.y = dev->size.y2;
 	drv->set_clip_area(dev, &dev->size);
@@ -681,7 +699,7 @@ void t_resize(struct graphics_device *dev)
 void t_kbd(struct graphics_device *dev, int key, int flags)
 {
 	struct terminal *term = dev->user_data;
-	struct event ev = { EV_KBD, 0, 0, 0 };
+	struct links_event ev = { EV_KBD, 0, 0, 0 };
 	struct rect r = {0, 0, 0, 0};
 	r.x2 = dev->size.x2, r.y2 = dev->size.y2;
 	ev.x = key;
@@ -703,7 +721,7 @@ void t_kbd(struct graphics_device *dev, int key, int flags)
 void t_mouse(struct graphics_device *dev, int x, int y, int b)
 {
 	struct terminal *term = dev->user_data;
-	struct event ev = { EV_MOUSE, 0, 0, 0 };
+	struct links_event ev = { EV_MOUSE, 0, 0, 0 };
 	struct rect r = {0, 0, 0, 0};
 	int bt, ac;
 	if (x == term->last_mouse_x && y == term->last_mouse_y && b == term->last_mouse_b) {
@@ -730,7 +748,7 @@ void t_mouse(struct graphics_device *dev, int x, int y, int b)
 
 static void in_term(struct terminal *term)
 {
-	struct event *ev;
+	struct links_event *ev;
 	int r;
 	unsigned char *iq;
 	NO_GFX;
@@ -745,41 +763,51 @@ static void in_term(struct terminal *term)
 	}
 	term->qlen += r;
 	test_queue:
-	if ((size_t)term->qlen < sizeof(struct event)) return;
-	ev = (struct event *)iq;
-	r = sizeof(struct event);
+	if ((size_t)term->qlen < sizeof(struct links_event)) return;
+	ev = (struct links_event *)iq;
+	r = sizeof(struct links_event);
 	if (ev->ev != EV_INIT && ev->ev != EV_RESIZE && ev->ev != EV_REDRAW && ev->ev != EV_KBD && ev->ev != EV_MOUSE && ev->ev != EV_ABORT) {
 		error("ERROR: error on terminal: bad event %d", ev->ev);
 		goto mm;
 	}
 	if (ev->ev == EV_INIT) {
 		int init_len;
-		if ((size_t)term->qlen < sizeof(struct event) + MAX_TERM_LEN + MAX_CWD_LEN + 2 * sizeof(int)) return;
-		init_len = *(int *)(iq + sizeof(struct event) + MAX_TERM_LEN + MAX_CWD_LEN + sizeof(int));
-		if ((size_t)term->qlen < sizeof(struct event) + MAX_TERM_LEN + MAX_CWD_LEN + 2 * sizeof(int) + init_len) return;
-		memcpy(term->term, iq + sizeof(struct event), MAX_TERM_LEN);
+		if ((size_t)term->qlen < sizeof(struct links_event) + MAX_TERM_LEN + MAX_CWD_LEN + 2 * sizeof(int)) return;
+		init_len = *(int *)(iq + sizeof(struct links_event) + MAX_TERM_LEN + MAX_CWD_LEN + sizeof(int));
+		if ((size_t)term->qlen < sizeof(struct links_event) + MAX_TERM_LEN + MAX_CWD_LEN + 2 * sizeof(int) + init_len) return;
+		memcpy(term->term, iq + sizeof(struct links_event), MAX_TERM_LEN);
 		term->term[MAX_TERM_LEN - 1] = 0;
-		memcpy(term->cwd, iq + sizeof(struct event) + MAX_TERM_LEN, MAX_CWD_LEN);
+		memcpy(term->cwd, iq + sizeof(struct links_event) + MAX_TERM_LEN, MAX_CWD_LEN);
 		term->cwd[MAX_CWD_LEN - 1] = 0;
-		term->environment = *(int *)(iq + sizeof(struct event) + MAX_TERM_LEN + MAX_CWD_LEN);
-		ev->b = (my_intptr_t)(iq + sizeof(struct event) + MAX_TERM_LEN + MAX_CWD_LEN + sizeof(int));
-		r = (int)sizeof(struct event) + MAX_TERM_LEN + MAX_CWD_LEN + 2 * (int)sizeof(int) + init_len;
+		term->environment = *(int *)(iq + sizeof(struct links_event) + MAX_TERM_LEN + MAX_CWD_LEN);
+		ev->b = (my_intptr_t)(iq + sizeof(struct links_event) + MAX_TERM_LEN + MAX_CWD_LEN + sizeof(int));
+		r = (int)sizeof(struct links_event) + MAX_TERM_LEN + MAX_CWD_LEN + 2 * (int)sizeof(int) + init_len;
 		sync_term_specs();
 	}
 	if (ev->ev == EV_REDRAW || ev->ev == EV_RESIZE || ev->ev == EV_INIT) {
 		struct window *win;
+
+		term->real_x = ev->x;
+		term->real_y = ev->y;
+
+		set_margin(term);
+
 		send_redraw:
 		if (ev->x < 0 || ev->y < 0) {
 			error("ERROR: bad terminal size: %d, %d", (int)ev->x, (int)ev->y);
 			goto mm;
 		}
-		alloc_term_screen(term, ev->x, ev->y);
+		alloc_term_screen(term);
 		clear_terminal(term);
 		erase_screen(term);
 		term->redrawing = 1;
 		foreachback(win, term->windows) win->handler(win, ev, 0);
 		term->redrawing = 0;
 	}
+	if (ev->ev == EV_MOUSE) {
+		ev->x -= term->left_margin;
+		ev->y -= term->top_margin;
+	}
 	if (ev->ev == EV_KBD || ev->ev == EV_MOUSE) {
 		if (ev->ev == EV_KBD && upcase(ev->x) == 'L' && ev->y & KBD_CTRL) {
 			ev->ev = EV_REDRAW;
@@ -836,7 +864,7 @@ static unsigned char frame_restrict[48] = {
 	179,  0,  0,  0,  0,  0,  0,  0,
 };
 
-#if defined(ENABLE_UTF8) && defined(WIN32)
+#if defined(ENABLE_UTF8) && defined(WIN)
 static inline char_t utf8_hack(char_t c)
 {
 	/*
@@ -862,9 +890,9 @@ static inline char_t utf8_hack(char_t c)
 #define SETPOS(x, y)							\
 {									\
 	add_to_str(&a, &l, cast_uchar "\033[");				\
-	add_num_to_str(&a, &l, (y) + 1);				\
+	add_num_to_str(&a, &l, (y) + 1 + term->top_margin);		\
 	add_to_str(&a, &l, cast_uchar ";");				\
-	add_num_to_str(&a, &l, (x) + 1);				\
+	add_num_to_str(&a, &l, (x) + 1 + term->left_margin);		\
 	add_to_str(&a, &l, cast_uchar "H");				\
 	n_chars = 0;							\
 }
@@ -910,7 +938,8 @@ static inline char_t utf8_hack(char_t c)
 		if (attrib & 0100) add_to_str(&a, &l, cast_uchar ";1");	\
 		add_to_str(&a, &l, cast_uchar "m");			\
 	}								\
-	if (c >= ' ' && c != 127 /*&& c != 155*/) {			\
+	if (c >= ' ' && c != 127 && (c != 155 ||			\
+	    (term->spec->charset != utf8_table && cp2u(155, term->spec->charset) != -1))) {\
 		if (c < 128 || frm || term->spec->charset != utf8_table) {\
 			add_chr_to_str(&a, &l, (unsigned char)c);	\
 		} else {						\
@@ -978,7 +1007,7 @@ void redraw_screen(struct terminal *term)
 			/*fprintf(stderr, "%d.%d : %d-%d -> %d-%d\n", x, y, term->last_screen[p].ch, term->last_screen[p].at, term->screen[p].ch, term->screen[p].at);*/
 			memcpy(&term->last_screen[p], &term->screen[p], sizeof(chr));
 #ifdef OPENVMS
-			if (n_chars >= 16) cy = -1;
+			if (n_chars >= term->x - 6) cy = -1;
 #endif
 			if (cx == x && cy == y) goto pc;/*PRINT_CHAR(p)*/
 			else if (cy == y && x - cx < 10 && x - cx > 0) {
@@ -1007,13 +1036,36 @@ void redraw_screen(struct terminal *term)
 		term->lcx = term->cx;
 		term->lcy = term->cy;
 		add_to_str(&a, &l, cast_uchar "\033[");
-		add_num_to_str(&a, &l, term->cy + 1);
+		add_num_to_str(&a, &l, term->cy + 1 + term->top_margin);
 		add_to_str(&a, &l, cast_uchar ";");
-		add_num_to_str(&a, &l, term->cx + 1);
+		add_num_to_str(&a, &l, term->cx + 1 + term->left_margin);
 		add_to_str(&a, &l, cast_uchar "H");
 	}
 	if (l && term->master) want_draw();
+#ifdef OPENVMS
+	{
+/*
+ * OpenVMS/VAX has some bug in the terminal driver and corrupts long strings.
+ * Also, we need to avoid breaking escape sequences.
+ */
+#define PRINT_BATCH	2000
+		int i, q;
+		for (i = 0; i < l; i += q) {
+			int qq;
+			q = PRINT_BATCH;
+			if (q > l - i) q = l - i;
+			for (qq = q - 1; qq > 0; qq--) {
+				if (a[i + qq] == 27) {
+					q = qq;
+					break;
+				}
+			}
+			hard_write(term->fdout, a + i, q);
+		}
+	}
+#else
 	hard_write(term->fdout, a, l);
+#endif
 	if (l && term->master) done_draw();
 	mem_free(a);
 	term->dirty = 0;
@@ -1026,16 +1078,13 @@ void destroy_terminal(struct terminal *term)
 	while ((term->windows.next) != &term->windows) delete_window(term->windows.next);
 	/*if (term->cwd) mem_free(term->cwd);*/
 	del_from_list(term);
-	if (term->blocked != -1) {
-		EINTRLOOP(rs, close(term->blocked));
-		set_handlers(term->blocked, NULL, NULL, NULL, NULL);
-	}
+	close_socket(&term->blocked);
 	if (term->title) mem_free(term->title);
 	if (!F) {
 		mem_free(term->screen);
 		mem_free(term->last_screen);
-		set_handlers(term->fdin, NULL, NULL, NULL, NULL);
 		mem_free(term->input_queue);
+		set_handlers(term->fdin, NULL, NULL, NULL);
 		EINTRLOOP(rs, close(term->fdin));
 		if (!term->master) {
 			if (term->fdout != term->fdin)
@@ -1052,6 +1101,10 @@ void destroy_terminal(struct terminal *term)
 		drv->shutdown_device(term->dev);
 #endif
 	}
+	if (term->handle_to_close != -1) {
+		hard_write(term->handle_to_close, cast_uchar "x", 1);
+		close_socket(&term->handle_to_close);
+	}
 	mem_free(term);
 	check_if_no_terminal();
 }
@@ -1218,9 +1271,7 @@ void exec_thread(unsigned char *path, int p)
 void close_handle(void *p)
 {
 	int h = (int)(my_intptr_t)p;
-	int rs;
-	EINTRLOOP(rs, close(h));
-	set_handlers(h, NULL, NULL, NULL, NULL);
+	close_socket(&h);
 }
 
 static void unblock_terminal(struct terminal *term)
@@ -1228,7 +1279,7 @@ static void unblock_terminal(struct terminal *term)
 	close_handle((void *)(my_intptr_t)term->blocked);
 	term->blocked = -1;
 	if (!F) {
-		set_handlers(term->fdin, (void (*)(void *))in_term, NULL, (void (*)(void *))destroy_terminal, term);
+		set_handlers(term->fdin, (void (*)(void *))in_term, NULL, term);
 		unblock_itrm(term->fdin);
 		/* clear the dirty flag because unblock_itrm queued a resize
 		   event - so avoid double redraw */
@@ -1292,7 +1343,10 @@ void exec_on_terminal(struct terminal *term, unsigned char *path, unsigned char
 				}
 #endif
 			}
-			if ((blockh = start_thread((void (*)(void *, int))exec_thread, param, (int)strlen(cast_const_char path) + (int)strlen(cast_const_char delet) + 3)) == -1) {
+#if defined(HAVE_MALLOC_TRIM)
+			malloc_trim(0);
+#endif
+			if ((blockh = start_thread((void (*)(void *, int))exec_thread, param, (int)strlen(cast_const_char path) + (int)strlen(cast_const_char delet) + 3, *delet != 0)) == -1) {
 				if (fg == 1) {
 					if (!F) unblock_itrm(term->fdin);
 #ifdef G
@@ -1312,12 +1366,12 @@ void exec_on_terminal(struct terminal *term, unsigned char *path, unsigned char
 #ifdef DOS
 				unblock_terminal(term);
 #else
-				set_handlers(blockh, (void (*)(void *))unblock_terminal, NULL, (void (*)(void *))unblock_terminal, term);
-				if (!F) set_handlers(term->fdin, NULL, NULL, (void (*)(void *))destroy_terminal, term);
+				set_handlers(blockh, (void (*)(void *))unblock_terminal, NULL, term);
+				if (!F) set_handlers(term->fdin, NULL, NULL, term);
 				/*block_itrm(term->fdin);*/
 #endif
 			} else {
-				set_handlers(blockh, close_handle, NULL, close_handle, (void *)(my_intptr_t)blockh);
+				set_handlers(blockh, close_handle, NULL, (void *)(my_intptr_t)blockh);
 			}
 		}
 	} else {
